import { build, createDevServer, createNitro } from 'nitropack';
import { toNodeListener } from 'h3';
import { normalizePath } from 'vite';
import { dirname, relative, resolve } from 'node:path';
import { platform } from 'node:os';
import { buildServer } from './build-server.js';
import { buildSSRApp } from './build-ssr.js';
import { pageEndpointsPlugin } from './plugins/page-endpoints.js';
import { getPageHandlers } from './utils/get-page-handlers.js';
import { buildSitemap } from './build-sitemap.js';
import { devServerPlugin } from './plugins/dev-server-plugin.js';
import { getMatchingContentFilesWithFrontMatter } from './utils/get-content-files.js';
const isWindows = platform() === 'win32';
let clientOutputPath = '';
const __dirname = dirname(new URL(import.meta.url).pathname);
export function nitro(options, nitroOptions) {
    const workspaceRoot = options?.workspaceRoot ?? process.cwd();
    const isTest = process.env['NODE_ENV'] === 'test' || !!process.env['VITEST'];
    const apiPrefix = `/${nitroOptions?.runtimeConfig?.['apiPrefix'] ?? 'api'}`;
    let isBuild = false;
    let isServe = false;
    let ssrBuild = false;
    let config;
    let nitroConfig;
    return [
        (options?.ssr
            ? devServerPlugin({
                entryServer: options?.entryServer,
                index: options?.index,
            })
            : false),
        {
            name: '@analogjs/vite-plugin-nitro',
            async config(_config, { command }) {
                isServe = command === 'serve';
                isBuild = command === 'build';
                ssrBuild = _config.build?.ssr === true;
                config = _config;
                const rootDir = relative(workspaceRoot, config.root || '.') || '.';
                const buildPreset = process.env['BUILD_PRESET'] ??
                    nitroOptions?.preset;
                const pageHandlers = getPageHandlers({ workspaceRoot, rootDir });
                nitroConfig = {
                    rootDir,
                    preset: buildPreset,
                    logLevel: nitroOptions?.logLevel || 0,
                    srcDir: normalizePath(`${rootDir}/src/server`),
                    scanDirs: [normalizePath(`${rootDir}/src/server`)],
                    output: {
                        dir: normalizePath(resolve(workspaceRoot, 'dist', rootDir, 'analog')),
                        publicDir: normalizePath(resolve(workspaceRoot, 'dist', rootDir, 'analog/public')),
                        ...nitroOptions?.output,
                    },
                    buildDir: normalizePath(resolve(workspaceRoot, 'dist', rootDir, '.nitro')),
                    typescript: {
                        generateTsConfig: false,
                    },
                    runtimeConfig: { ...nitroOptions?.runtimeConfig },
                    rollupConfig: {
                        onwarn(warning) {
                            if (warning.message.includes('empty chunk') &&
                                warning.message.endsWith('.server')) {
                                return;
                            }
                        },
                        plugins: [pageEndpointsPlugin()],
                    },
                    handlers: [
                        {
                            handler: normalizePath(`${__dirname}/runtime/api-middleware`),
                            middleware: true,
                        },
                        ...pageHandlers,
                    ],
                };
                if (isVercelPreset(buildPreset)) {
                    nitroConfig = withVercelOutputAPI(nitroConfig, workspaceRoot);
                }
                if (!ssrBuild && !isTest) {
                    // store the client output path for the SSR build config
                    clientOutputPath = resolve(rootDir, config.build?.outDir || 'dist/client');
                }
                nitroConfig.alias = {
                    // This is not the final fix but start point to discuss a fix for windows
                    '#analog/ssr': (isWindows ? 'file://' : '') +
                        normalizePath(resolve(workspaceRoot, 'dist', rootDir, 'ssr/main.server')),
                    '#analog/index': normalizePath(resolve(clientOutputPath, 'index.html')),
                    ...nitroOptions?.alias,
                };
                if (isBuild) {
                    if (isEmptyPrerenderRoutes(options)) {
                        nitroConfig.prerender = {};
                        nitroConfig.prerender.routes = ['/'];
                    }
                    if (options?.prerender) {
                        nitroConfig.prerender = nitroConfig.prerender ?? {};
                        nitroConfig.prerender.crawlLinks = options?.prerender?.discover;
                        let routes = [];
                        const prerenderRoutes = options?.prerender?.routes;
                        if (isArrayWithElements(prerenderRoutes)) {
                            routes = prerenderRoutes;
                        }
                        else if (typeof prerenderRoutes === 'function') {
                            routes = await prerenderRoutes();
                        }
                        nitroConfig.prerender.routes = routes.reduce((prev, current) => {
                            if (!current) {
                                return prev;
                            }
                            if (typeof current === 'string') {
                                prev.push(current);
                                return prev;
                            }
                            const affectedFiles = getMatchingContentFilesWithFrontMatter(workspaceRoot, rootDir, current.contentDir);
                            affectedFiles.forEach((f) => {
                                const result = current.transform(f);
                                if (result) {
                                    prev.push(result);
                                }
                            });
                            return prev;
                        }, []);
                    }
                    if (ssrBuild) {
                        nitroConfig = {
                            ...nitroConfig,
                            publicAssets: [{ dir: clientOutputPath }],
                            serverAssets: [
                                {
                                    baseName: 'public',
                                    dir: clientOutputPath,
                                },
                            ],
                            externals: {
                                ...nitroOptions?.externals,
                                external: [
                                    'rxjs',
                                    'node-fetch-native/dist/polyfill',
                                    ...(nitroOptions?.externals?.external || []),
                                ],
                            },
                            moduleSideEffects: [
                                'zone.js/node',
                                'zone.js/fesm2015/zone-node',
                                ...(nitroOptions?.moduleSideEffects || []),
                            ],
                            renderer: normalizePath(`${__dirname}/runtime/renderer`),
                            handlers: [
                                {
                                    handler: normalizePath(`${__dirname}/runtime/api-middleware`),
                                    middleware: true,
                                },
                                ...pageHandlers,
                            ],
                        };
                    }
                }
                nitroConfig = {
                    ...nitroConfig,
                    ...nitroOptions,
                };
            },
            async configureServer(viteServer) {
                if (isServe && !isTest) {
                    const nitro = await createNitro({
                        dev: true,
                        ...nitroConfig,
                    });
                    const server = createDevServer(nitro);
                    await build(nitro);
                    viteServer.middlewares.use(apiPrefix, toNodeListener(server.app));
                    viteServer.httpServer?.once('listening', () => {
                        process.env['ANALOG_HOST'] = !viteServer.config.server.host
                            ? 'localhost'
                            : viteServer.config.server.host;
                        process.env['ANALOG_PORT'] = `${viteServer.config.server.port}`;
                    });
                    console.log(`\n\nThe server endpoints are accessible under the "${apiPrefix}" path.`);
                }
            },
            async closeBundle() {
                if (ssrBuild) {
                    return;
                }
                if (isBuild) {
                    if (options?.ssr) {
                        console.log('Building SSR application...');
                        await buildSSRApp(config, options);
                    }
                    await buildServer(options, nitroConfig);
                    if (nitroConfig.prerender?.routes?.length &&
                        options?.prerender?.sitemap) {
                        console.log('Building Sitemap...');
                        // sitemap needs to be built after all directories are built
                        await buildSitemap(config, options.prerender.sitemap, nitroConfig.prerender.routes, nitroConfig.output?.publicDir);
                    }
                    console.log(`\n\nThe '@analogjs/platform' server has been successfully built.`);
                }
            },
        },
    ];
}
function isEmptyPrerenderRoutes(options) {
    if (!options || isArrayWithElements(options?.prerender?.routes)) {
        return false;
    }
    return !options.prerender?.routes;
}
function isArrayWithElements(arr) {
    return !!(Array.isArray(arr) && arr.length);
}
const isVercelPreset = (buildPreset) => process.env['VERCEL'] ||
    (buildPreset && buildPreset.toLowerCase().includes('vercel'));
const withVercelOutputAPI = (nitroConfig, workspaceRoot) => ({
    ...nitroConfig,
    output: {
        ...nitroConfig?.output,
        dir: normalizePath(resolve(workspaceRoot, '.vercel', 'output')),
        publicDir: normalizePath(resolve(workspaceRoot, '.vercel', 'output/static')),
    },
});
//# sourceMappingURL=vite-plugin-nitro.js.map